#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from os import PathLike
from pathlib import Path
from fhir.resources.R4B import bundle
from pydicom import dcmread
from pydicom import dataset
from tqdm import tqdm
import logging
from typing import Iterable, Union
import dicom2fhir.helpers as helpers
from dicom2fhir.helpers import get_or
from dicom2fhir.dicom2fhirbundle import Dicom2FHIRBundle

StrPath = Union[str, PathLike]

def _parse_directory(dcmDir: StrPath, config: dict) -> Iterable[dataset.Dataset]:
    """
    Parse a directory of DICOM files including subdirectories and return instances as a Generator.

    :param dcmDir: Directory containing DICOM files.
    :return: Iterable[dataset.Dataset]
    """
    base = Path(dcmDir)
    if not base.is_dir():
        raise ValueError(f"Directory '{dcmDir}' not found")

    skip_invalid_files = get_or(config, "directory_parser.skip_invalid_files", True)

    def is_dicom_file(path: str) -> bool:
        try:
            with open(path, 'rb') as f:
                f.seek(128)
                return f.read(4) == b'DICM'
        except Exception:
            return False

    for fp in tqdm(base.rglob("*")):

        if not fp.is_file():
            continue
        if skip_invalid_files and not is_dicom_file(str(fp)):
            logging.warning(f"Skipping invalid DICOM file: {fp}")
            continue

        try:
            yield dcmread(str(fp), stop_before_pixels=True, force=True)
        except:
            logging.exception(f"An error occurred while processing DICOM file {fp}")
            raise

def _create_bundle(instances: Iterable[dataset.Dataset], config: dict = {}) -> bundle.Bundle:

    dcm2fhir = Dicom2FHIRBundle(config=config)
    
    for ds in instances:
        if not isinstance(ds, dataset.Dataset):
            raise TypeError("Expected a pydicom Dataset object")
        dcm2fhir.add(ds)

    return dcm2fhir.create_bundle()

def process_dicom_2_fhir(dcms: StrPath | Iterable[dataset.Dataset], config: dict = {}) -> bundle.Bundle:
    """
    Process DICOM files or datasets into an ImagingStudy FHIR resource.
    
    :param dcms: Either a directory containing DICOM files or an iterable of DICOM datasets.
    :return: ImagingStudy resource.
    """

    # use default id function for FHIR resource id generation
    if 'id_function' not in config:
        config['id_function'] = helpers.default_id_function()

    instances = _parse_directory(dcms, config) if isinstance(dcms, StrPath) else dcms
    return _create_bundle(instances, config)



# dicom2device.py

import uuid
from collections.abc import Iterable
from pydicom.dataset import Dataset
from fhir.resources.R4B.device import Device, DeviceDeviceName
from fhir.resources.R4B.annotation import Annotation

def _map_software_versions(ds: Dataset) -> list[dict]:
    """
    Extract SoftwareVersions (DICOM 0018,1020) from the dataset and map to FHIR Device.version.
    
    This attribute is multi-valued (LO), so it returns a list of version strings.
    """
    # pydicom returns either a single value or a MultiValue object for VR LO
    raw = ds.get("SoftwareVersions", None)
    if not raw:
        return []

    # Normalize to list of strings
    if isinstance(raw, Iterable) and not isinstance(raw, (str, bytes)):
        return [{'value': str(item).strip()} for item in raw if item is not None]
    else:
        return [{'value': str(raw).strip()}]

def build_device_resource(ds: Dataset, config: dict) -> Device:
    """
    Build FHIR Device resource from DICOM metadata (General Equipment Module).
    Extracts manufacturer, model, serial, version, institution, station, calibration, UDI, etc.
    """

    device = Device.model_construct()
    # Resource ID
    id_func = config.get("id_function", lambda t, d: str(uuid.uuid4()))
    device.id = id_func("Device", ds)

    # Identifiers
    identifiers = []
    for tag_name, system in [
        ("DeviceSerialNumber", "urn:dicom:device-serial-number"),
        ("DeviceUID", "urn:dicom:device-uid"),
    ]:
        val = ds.get(tag_name)
        if val:
            identifiers.append({"use": "official", "system": system, "value": val})
    if identifiers:
        device.identifier = identifiers

    # Manufacturer & Model
    if (m := ds.get("Manufacturer")):
        device.manufacturer = m
    if (name := ds.get("ManufacturerModelName")):
        device.deviceName = [DeviceDeviceName.model_construct(name=name, type="model-name")]

    # Software version(s)
    device.version = _map_software_versions(ds)

    # Institutional context
    for field, tag in [
        ("manufacturer", "Manufacturer"),
        ("owner", ds.get("InstitutionName")),
    ]:
        pass  # Use owner assignment below
    if (inst := ds.get("InstitutionName")):
        device.owner = {"display": inst}
    if (dept := ds.get("InstitutionalDepartmentName")):
        device.location = {"display": dept}
    if (station := ds.get("StationName")):
        # Could be assigned to device.deviceName as 'station' or use part of location
        device.deviceName = device.deviceName or []
        device.deviceName.append(DeviceName.model_construct(name=station, type="station-name"))

    # Physical/device-specific details
    if (spat := ds.get("SpatialResolution")):
        device.property = device.property or []
        device.property.append({
            "type": {"text": "spatial-resolution-mm"},
            "valueQuantity": {"value": float(spat), "unit": "mm"}
        })

    # Calibration date/time
    cal_date = ds.get("DateOfLastCalibration")
    cal_time = ds.get("TimeOfLastCalibration")
    if cal_date or cal_time:
        dt = cal_date + (cal_time or "")
        device.note = device.note or []
        device.note.append(Annotation.model_construct(text=f"Last calibration: {dt}"))

    # Pixel padding—maybe not core but included
    if (pad := ds.get("PixelPaddingValue")):
        device.note = device.note or []
        device.note.append(Annotation.model_construct(text=f"Pixel padding value: {pad}"))

    # UDI (Unique Device Identifier)
    if hasattr(ds, "UDISequence"):
        udi_items = []
        for item in ds.UDISequence:
            if (code := ds.get("DeviceIdentifier")):
                udi_items.append({"system": "http://hl7.org/fhir/sid/udi", "value": code})
        if udi_items:
            device.udiCarrier = udi_items

    # Modality as device type
    if (mod := ds.get("Modality")):
        device.type = {"text": mod}

    return device


from datetime import datetime
from dateutil import tz as dateutil_tz

from fhir.resources.R4B import imagingstudy
from fhir.resources.R4B import identifier
from fhir.resources.R4B import codeableconcept
from fhir.resources.R4B import coding
from fhir.resources.R4B import patient
from fhir.resources.R4B import humanname
from fhir.resources.R4B import fhirtypes
from fhir.resources.R4B import reference
from fhir.resources.R4B import extension
import pandas as pd
import json
from pathlib import Path
import os
import logging

TERMINOLOGY_CODING_SYS = "http://terminology.hl7.org/CodeSystem/v2-0203"
TERMINOLOGY_CODING_SYS_CODE_ACCESSION = "ACSN"
TERMINOLOGY_CODING_SYS_CODE_MRN = "MR"

ACQUISITION_MODALITY_SYS = "http://dicom.nema.org/resources/ontology/DCM"
SCANNING_SEQUENCE_SYS = "https://dicom.nema.org/medical/dicom/current/output/chtml/part03/sect_C.8.3.html"
SCANNING_VARIANT_SYS = "https://dicom.nema.org/medical/dicom/current/output/chtml/part03/sect_C.8.3.html"

SOP_CLASS_SYS = "urn:ietf:rfc:3986"

# load rather expesive resource into global var to make it reusable
BODYSITE_SNOMED_MAPPING_PATH = Path(__file__).parent / "resources" / "terminologies" / "bodysite_snomed.json"
BODYSITE_SNOMED_MAPPING = pd.DataFrame(json.loads(BODYSITE_SNOMED_MAPPING_PATH.read_text(encoding="utf-8")))

def _get_snomed(dicom_bodypart: str, sctmapping: pd.DataFrame) -> dict[str, str] | None:
    _rec = sctmapping.loc[sctmapping['Body Part Examined'] == dicom_bodypart]
    if _rec.empty:
        return None
    return {
        'code': _rec["Code Value"].iloc[0],
        'display': _rec["Code Meaning"].iloc[0],
    }

def gen_accession_identifier(id):
    idf = identifier.Identifier.model_construct()
    idf.use = "usual"
    idf.type = codeableconcept.CodeableConcept.model_construct()
    idf.type.coding = []
    acsn = coding.Coding.model_construct()
    acsn.system = TERMINOLOGY_CODING_SYS
    acsn.code = TERMINOLOGY_CODING_SYS_CODE_ACCESSION

    idf.type.coding.append(acsn)
    idf.value = id
    return idf


def gen_studyinstanceuid_identifier(id):
    idf = identifier.Identifier.model_construct()
    idf.system = "urn:dicom:uid"
    idf.value = "urn:oid:" + id
    return idf


def get_patient_resource_ids(PatientID, IssuerOfPatientID):
    idf = identifier.Identifier.model_construct()
    idf.use = "usual"
    idf.value = PatientID

    idf.type = codeableconcept.CodeableConcept.model_construct()
    idf.type.coding = []
    id_coding = coding.Coding.model_construct()
    id_coding.system = TERMINOLOGY_CODING_SYS
    id_coding.code = TERMINOLOGY_CODING_SYS_CODE_MRN
    idf.type.coding.append(id_coding)

    if IssuerOfPatientID is not None:
        idf.assigner = reference.Reference.model_construct()
        idf.assigner.display = IssuerOfPatientID

    return idf


def calc_gender(gender):
    if gender is None:
        return "unknown"
    if not gender:
        return "unknown"
    if gender.upper().lower() == "f":
        return "female"
    if gender.upper().lower() == "m":
        return "male"
    if gender.upper().lower() == "o":
        return "other"

    return "unknown"


def calc_dob(dicom_dob):
    if dicom_dob == '':
        return None

    try:
        dob = datetime.strptime(dicom_dob, '%Y%m%d')
        fhir_dob = fhirtypes.Date(
            dob.year,
            dob.month,
            dob.day
        )
    except Exception:
        return None
    return fhir_dob


def inline_patient_resource(referenceId, PatientID, IssuerOfPatientID, patientName, gender, dob):
    p = patient.Patient.model_construct()
    p.id = referenceId
    p.name = []
    # p.use = "official"
    p.identifier = [get_patient_resource_ids(PatientID, IssuerOfPatientID)]
    hn = humanname.HumanName.model_construct()
    hn.family = patientName.family_name
    if patientName.given_name != '':
        hn.given = [patientName.given_name]
    p.name.append(hn)
    p.gender = calc_gender(gender)
    p.birthDate = calc_dob(dob)
    p.active = True
    return p


def gen_procedurecode_array(procedures):
    if procedures is None:
        return None
    fhir_proc = []
    for p in procedures:
        concept = codeableconcept.CodeableConcept.model_construct()
        c = coding.Coding.model_construct()
        c.system = p["system"]
        c.code = p["code"]
        c.display = p["display"]
        concept.coding = []
        concept.coding.append(c)
        concept.text = p["display"]
        fhir_proc.append(concept)
    if len(fhir_proc) > 0:
        return fhir_proc
    return None


def gen_started_datetime(dt, tm, tz):
    """
    Generate a timezone-aware datetime object from DICOM date and time strings.

    Args:
        dt (str): DICOM date in the format 'YYYYMMDD'.
        tm (str): DICOM time in the format 'HHMMSS' or shorter.
        tz (str): Timezone as a string (e.g., 'Europe/Berlin' or '+01:00').

    Returns:
        datetime: A timezone-aware datetime object or None.
    """
    if dt is None:
        return None

    dt_pattern = '%Y%m%d'
    if tm is not None and len(tm) >= 6:
        studytm = datetime.strptime(tm[0:6], '%H%M%S')
        dt_string = f"{dt} {studytm.hour:02d}:{studytm.minute:02d}:{studytm.second:02d}"
        dt_pattern += " %H:%M:%S"
    else:
        dt_string = dt

    try:
        dt_date = datetime.strptime(dt_string, dt_pattern)
    except ValueError:
        return None

    # Apply timezone
    try:
        if tz:
            tzinfo = dateutil_tz.gettz(tz)
            if tzinfo is not None:
                dt_date = dt_date.replace(tzinfo=tzinfo)
    except Exception:
        pass

    return dt_date

def gen_reason(reason, reasonStr):
    if reason is None and reasonStr is None:
        return None
    reasonList = []
    if reason is None or len(reason) <= 0:
        # Only assign if non-empty and not just whitespace
        if reasonStr and reasonStr.strip():
            rc = codeableconcept.CodeableConcept.model_construct()
            rc.text = reasonStr
            reasonList.append(rc)
        return reasonList

    for r in reason:
        rc = codeableconcept.CodeableConcept.model_construct()
        rc.coding = []
        c = coding.Coding.model_construct()
        c.system = r["system"]
        c.code = r["code"]
        c.display = r["display"]
        rc.coding.append(c)
        reasonList.append(rc)
    return reasonList


def gen_coding(code: str, system: str|None = None, display: str|None = None):
    if isinstance(code, list):
        raise Exception(
        "More than one code for type Coding detected")
    c = coding.Coding.model_construct()
    c.code = code
    c.system = system
    c.display = display
    if system is None and display is None:
        c.userSelected = True

    return c

def gen_codeable_concept(value_list: list, system):
    c = codeableconcept.CodeableConcept.model_construct()
    c.coding = []
    for _l in value_list:
        m = gen_coding(_l, system)
        c.coding.append(m)
    return c

def gen_bodysite_coding(bd):

    bd_snomed = _get_snomed(bd, sctmapping=BODYSITE_SNOMED_MAPPING)
    
    if bd_snomed is None:
        return gen_coding(code=str(bd))
    
    return gen_coding(
        code=str(bd_snomed['code']),
        system="http://snomed.info/sct",
        display=bd_snomed['display']
    ) 

# def update_study_modality_list(study_list_modality: list, modality: str):
#     if study_list_modality is None or len(study_list_modality) <= 0:
#         study_list_modality = []
#         study_list_modality.append(modality)
#         return

#     c = next((mc for mc in study_list_modality if
#               mc == modality), None)
#     if c is not None:
#         return

#     study_list_modality.append(modality)
#     return


# def update_study_bodysite_list(study: imagingstudy.ImagingStudy, bodysite: coding.Coding):
#     if study.bodySite__ext is None or len(study.bodySite__ext) <= 0:
#         study.bodySite__ext = []
#         study.bodySite__ext.append(bodysite)
#         return

#     c = next((mc for mc in study.bodySite__ext if
#               mc.system == bodysite.system and
#               mc.code == bodysite.code), None)
#     if c is not None:
#         return

#     study.bodySite__ext.append(bodysite)
#     return


# def update_study_laterality_list(study: imagingstudy.ImagingStudy, laterality: coding.Coding):
#     if study.laterality__ext is None or len(study.laterality__ext) <= 0:
#         study.laterality__ext = []
#         study.laterality__ext.append(laterality)
#         return

#     c = next((mc for mc in study.laterality__ext if
#               mc.system == laterality.system and
#               mc.code == laterality.code), None)
#     if c is not None:
#         return

    # study.laterality__ext.append(laterality)
    # return


def dcm_coded_concept(CodeSequence):
    concepts = []
    for seq in CodeSequence:
        concept = {}
        concept["code"] = seq[0x0008, 0x0100].value
        concept["system"] = seq[0x0008, 0x0102].value
        concept["display"] = seq[0x0008, 0x0104].value
        concepts.append(concept)
    return concepts


import uuid
from fhir.resources.R4B import bundle
from fhir.resources.R4B import imagingstudy
from fhir.resources.R4B import patient
from fhir.resources.R4B import device
from fhir.resources.R4B.reference import Reference
from pydicom import dataset
import logging
from dicom2fhir.dicom2fhirutils import gen_coding, SOP_CLASS_SYS, ACQUISITION_MODALITY_SYS, gen_bodysite_coding, gen_accession_identifier, gen_studyinstanceuid_identifier, dcm_coded_concept, gen_procedurecode_array, gen_started_datetime, gen_reason
from dicom2fhir.dicom2patient import build_patient_resource
from dicom2fhir.dicom2observation import build_observation_resources
from dicom2fhir.dicom2device import build_device_resource
from dicom2fhir.helpers import get_or
class Dicom2FHIRBundle():

    def __init__(self, config: dict = {}):
        """
        Initialize the Dicom2FHIRBundle with an optional configuration.
        """
        self.study: imagingstudy.ImagingStudy | None = None
        self.series = {}
        self.instances = {}
        # Device
        self.device: device.Device | None = None
        # Patient
        self.pat: patient.Patient | None = None
        self.obs = []
        self.config = config
        
    def add(self, ds: dataset.Dataset):
        """
        Add a DICOM dataset to the ImagingStudy.
        """
        if not isinstance(ds, dataset.Dataset):
            raise TypeError("Expected a pydicom Dataset object")

        # is first instance?
        if self.study is None:
            self.device = build_device_resource(ds, self.config)
            self.pat = build_patient_resource(ds, self.config)
            self._create_imaging_study(ds)
            if get_or(self.config, "generator.observation.add_vital_signs", True):
                self.obs = build_observation_resources(ds, self.pat, self.study, self.config) or []

        self._add_imaging_study_series(ds)
        self._add_instance(ds)

    def _create_imaging_study(self, ds):

        study_data = {}
        study_data["resource_type"] = "ImagingStudy"
        study_data["id"] = self.config['id_function']("ImagingStudy", ds)
        study_data["status"] = "available"
        try:
            if ds.StudyDescription != '':
                study_data["description"] = ds.StudyDescription
        except Exception:
            pass  # missing study description

        study_data["identifier"] = []
        study_data["identifier"].append(gen_accession_identifier(ds.AccessionNumber))
        study_data["identifier"].append(gen_studyinstanceuid_identifier(ds.StudyInstanceUID))

        # Set the patient reference
        study_data["subject"] = Reference.model_construct(reference=f"Patient/{self.pat.id}") if self.pat else None

        procedures = []
        try:
            procedures = dcm_coded_concept(ds.ProcedureCodeSequence)
        except Exception:
            pass  # procedure code sequence not found

        study_data["procedureCode"] = gen_procedurecode_array(procedures)

        studyTime = None
        try:
            studyTime = ds.StudyTime
        except Exception:
            pass  # print("Study Date is missing")

        try:
            studyDate = ds.StudyDate
            study_data["started"] = gen_started_datetime(studyDate, studyTime, self.config["dicom_timezone"])
        except Exception:
            pass  # print("Study Date is missing")

        reason = None
        reasonStr = None
        try:
            reason = dcm_coded_concept(ds.ReasonForRequestedProcedureCodeSequence)
        except Exception:
            pass  # print("Reason for Request procedure Code Seq is not available")

        try:
            reasonStr = ds.ReasonForTheRequestedProcedure
        except Exception:
            pass  # print ("Reason for Requested procedures not found")

        study_data["reasonCode"] = gen_reason(reason, reasonStr)

        study_data["numberOfSeries"] = 0
        study_data["numberOfInstances"] = 0

        # instantiate study here, when all required fields are available
        self.study = imagingstudy.ImagingStudy(**study_data)
        
    def _add_imaging_study_series(self, ds: dataset.Dataset):

        series_instance_uid = str(ds.SeriesInstanceUID)

        if series_instance_uid in self.series:
            return  # Series already exists

        # inti data container
        self.series[series_instance_uid] = {}

        
        self.series[series_instance_uid]["uid"] = series_instance_uid
        try:
            if ds.SeriesDescription != '':
                self.series[series_instance_uid]["description"] = ds.SeriesDescription
        except Exception:
            pass

        self.series[series_instance_uid]["number"] = ds.SeriesNumber
        self.series[series_instance_uid]["numberOfInstances"] = 0

        self.series[series_instance_uid]["modality"] = gen_coding(
            code=ds.Modality,
            system=ACQUISITION_MODALITY_SYS
        )
        
        try:
            stime = ds.SeriesTime
            sdate = ds.SeriesDate
            self.series[series_instance_uid]["started"] = gen_started_datetime(sdate, stime, self.config["dicom_timezone"])
        except Exception:
            pass  # print("Series Date/Time is missing")

        try:
            self.series[series_instance_uid]["bodySite"] = gen_bodysite_coding(ds.BodyPartExamined)
        except Exception:
            pass  # print ("Body Part Examined missing")

        try:
            self.series[series_instance_uid]["laterality"] = gen_coding(ds.Laterality)
        except Exception:
            pass  # print ("Laterality missing")
    
    def _add_instance(self, ds: dataset.Dataset):

        series_instance_uid = str(ds.SeriesInstanceUID)
        sop_instance_uid = str(ds.SOPInstanceUID)

        if series_instance_uid not in self.instances:
            self.instances[series_instance_uid] = {}

        if sop_instance_uid in self.instances[series_instance_uid]:
            print("Error: SOP Instance UID already exists in the series")
            print(self.instances[series_instance_uid][sop_instance_uid].as_json())
            return

        self.instances[series_instance_uid][sop_instance_uid] = {}
       
        self.instances[series_instance_uid][sop_instance_uid]["uid"] = sop_instance_uid
        self.instances[series_instance_uid][sop_instance_uid]["sopClass"] = gen_coding(
            code="urn:oid:" + ds.SOPClassUID,
            system=SOP_CLASS_SYS
        )
        self.instances[series_instance_uid][sop_instance_uid]["number"] = ds.InstanceNumber

        try:
            if str(ds.Modality) == "SR":
                seq = ds.ConceptNameCodeSequence
                self.instances[series_instance_uid][sop_instance_uid]["title"] = seq[0x0008, 0x0104]
            else:
                self.instances[series_instance_uid][sop_instance_uid]["title"] = '\\'.join(ds.ImageType)
        except Exception:
            pass  # print("Unable to set instance title")

    def _build_imaging_study(self) -> imagingstudy.ImagingStudy:
        """
        Build the ImagingStudy resource from the collected data.
        """
        if self.study is None:
            raise ValueError("No ImagingStudy data has been added")

        # Add series to the study
        self.study.series = []
        for series_uid, series_data in self.series.items():

            # only add sop instance data if configured
            if get_or(self.config, "generator.imaging_study.add_instances", False):
                series_data["instance"] = []
                for instance_uid, instance_data in self.instances.get(series_uid, {}).items():
                    series_data["instance"].append(imagingstudy.ImagingStudySeriesInstance(**instance_data))
                series_data["numberOfInstances"] = len(series_data["instance"])
            else:
                series_data["numberOfInstances"] = len(self.instances[series_uid])

            # Create ImagingStudySeries object
            series = imagingstudy.ImagingStudySeries(**series_data)
            self.study.series.append(series)

        # Set the number of series and instances
        self.study.numberOfSeries = len(self.study.series)
        self.study.numberOfInstances = sum(s.numberOfInstances for s in self.study.series)

        # Set the modalities
        modality_set = {
            s.modality.code: s.modality
            for s in self.study.series or []
            if s.modality is not None
        }
        self.study.modality = list(modality_set.values())

        return self.study
    
    def create_bundle(self) -> bundle.Bundle:
        """
        Create the final transaction bundle.
        """

        def _to_entry(resource):
            return {
                #'fullUrl': f"urn:uuid:{resource.id}",
                'resource': resource,
                'request': {
                    'method': 'PUT',
                    'url': f"{resource.__resource_type__}/{resource.id}"
                }
            }

        if not self.study:
            raise ValueError("No ImagingStudy data has been added")

        # Build the ImagingStudy resource
        _study = self._build_imaging_study()
    
        # wrap entries in a transaction Bundle and return
        return bundle.Bundle.model_validate({
            'resourceType': 'Bundle',
            'type': "transaction",
            'id': str(uuid.uuid4()),
            'entry': [
                _to_entry(_study),
                _to_entry(self.pat),
                _to_entry(self.device)
            ] + [_to_entry(o) for o in self.obs]
        })